//algo.cpp
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <alloc.h>
//#include <GL/gl.h>
#include <math.h>
#include <pthread.h>
#include "algo.h"
#include "xtree.h"
#include "synchronizer.h"


tArretes * ARPM_Prim(EuclidianGraphe *e,unsigned long Options)
{
  OutputLevel Output=(OutputLevel)OPTION_OUTPUT(Options);
  if (Output>=Important) printf("Prim: gnration d'un arbre recouvrant de poid minimal\n");
  char *sommetscouverts=(char *)malloc(e->GetSommets());
  memset(sommetscouverts,0,e->GetSommets());
  sommetscouverts[0]=1;
  unsigned int meilleurs1,meilleurs2;
  unsigned long meilleurpoid;
  tArretes *ta=new tArretes(e->GetSommets());
  for (int i=1;i<e->GetSommets();i++)
    {//recherche de l'arrete la + courte reliant les sommets couverts aux autres
      tArrete *ar=NULL;
      meilleurpoid=0x7fffffff;
      for (int j=0;j<e->GetSommets()-1;j++)
	{
	  for (int k=j+1;k<e->GetSommets();k++)
	    {
	      if (sommetscouverts[j]+sommetscouverts[k]!=1) continue;
	      if (meilleurpoid>e->Get(j,k))
		{
		  meilleurpoid=e->Get(j,k);
		  meilleurs1=j;
		  meilleurs2=k;
		}
	    }
	}
      //on insere le meilleur
      if (Output>=Detailed) printf("insertion de %d-%d\n",meilleurs1,meilleurs2);
      ta->Add(meilleurs1,meilleurs2);
      sommetscouverts[meilleurs1]=1;
      sommetscouverts[meilleurs2]=1;//une des deux insertion est superflue, mais on evite un test
      if (OPTION_DRAW(Options)) 
	{
	  usleep(OPTION_DRAWWAIT(Options)*10000);
	  e->Draw(ta);
	}
    }
  free(sommetscouverts);
  return ta;
}

void gen_parcours_prefixe(tArretes *ta,unsigned long * arretes,unsigned long *pos,unsigned long sommet,unsigned long pere,bool output)
{
  tArrete *ar=NULL;
  int s;
  arretes[(*pos)]=sommet;
  (*pos)++;
  if (output) printf("gen_parcours_prefixe: ajout de %d, depuis %d\n",sommet,pere);
  while ((s=ta->Get(&ar,sommet))!=-1)
    {
      if (s==pere) continue;
      gen_parcours_prefixe(ta,arretes,pos,s,sommet,output);
    }
}

tArretes * PVC_HARPM(EuclidianGraphe *e,unsigned long Options)
{
  OutputLevel Output=(OutputLevel)OPTION_OUTPUT(Options);
  if (Output>=Important) printf("HARPM: parcours prefixe d'un ARPM sur un graphe de %d sommets\n",e->GetSommets());
  //on part d'un ARPM  
  tArretes *ta=ARPM_Prim(e,Options);
  //generation de la liste des sommets: parcours prefixe
  //le plus simple est d'utiliser une implementation recursive
  //une implementation itrative devrait de toute manire utiliser un systeme de pile
  unsigned long * arretes=(unsigned long *)malloc(e->GetSommets()*sizeof(unsigned long));
  unsigned long pos=0;
  gen_parcours_prefixe(ta,arretes,&pos,0,-1,Output>=Detailed);
  if (pos!=e->GetSommets())
    {
      if (Output>=Important) printf("HARPM: erreur, parcours prefixe: sommets manquants\n"); 
      return NULL;
    }
  //voila! plus qu'a remplir la structure tArretes
  ta->Clear();
  for (int i=0;i<e->GetSommets()-1;i++)
    ta->Add(arretes[i],arretes[i+1]);
  ta->Add(arretes[0],arretes[e->GetSommets()-1]);
  free(arretes);
  if (OPTION_DRAW(Options)) 
    {
      e->Draw(ta);
    }
 if (Output>=Important) printf("HARPM: fin, chemin de longueur %d trouve\n",e->Eval(ta));
 return ta;
}

tArretes * PVC_HMoindreCout(EuclidianGraphe *e,unsigned long Options)
{
  OutputLevel Output=(OutputLevel)OPTION_OUTPUT(Options);
  if (Output>=Important) printf("HMoindreCout: Heuristique du moindre cout sur un graphe de %d sommets\n",e->GetSommets());
  tArretes * circuit=new tArretes(e->GetSommets());
  unsigned long MeilleurSommet; //sommet a rajouter diminuant au plus le coup
  long MeilleurCircuit; //stoque sous forme d'un delta avec le circuit actuel
  unsigned long PoidCircuit; //poid du circuit actuel
  char * SommetsTraites=new char[e->GetSommets()]; //sommets deja dans le circuit
  tArrete * Circuit=NULL;//le circuit sous forme d'une liste chainee circulaire (pointeurs Next2 et Prev2), respectant la regle this->s1=this->Prev2->s2
  tArrete * Place; //lieu de la meilleure insertion
  memset(SommetsTraites,0,e->GetSommets());
  //initialisation: on part d'un sommet
  SommetsTraites[0]=1;
  PoidCircuit=0;
  //recurence  
//on ajoute le sommet rallongeant le moins le circuit
  for (int k=1;k<e->GetSommets();k++)
    {
      if (Output>=UltraDetailed) printf ("HMoindreCout: recherche du %d ieme sommet a inserer\n",k+1);
      MeilleurCircuit=0x7fffffff; //initialisation a valeur elevee
      //on teste chaque sommet
      for (int j=0;j<e->GetSommets();j++)
	{
	  if (SommetsTraites[j]) continue; //sommet deja dans le circuit
	  //printf("Check sommet %d\n",j); 
	  //cas particulier du premier sommet
	  if (!Circuit)
	    {
	      unsigned long l=e->Get(0,j);
	      if (MeilleurCircuit>l)
		{
		  MeilleurCircuit=l; //l*2 techniquement
		  MeilleurSommet=j;
		}
	    }
	  else
	    {
	      //on teste tous les points d'insertion, a savoir toutes les arretes de circuit
	      tArrete * a=Circuit;
	      do
		{
		  if (Output>=UltraDetailed) printf("Check %d-%d\n",a->s1,a->s2);
		  unsigned long l=e->Get(j,a->s1)+e->Get(j,a->s2)-e->Get(a->s1,a->s2);
		  if (MeilleurCircuit>l)
		    {
		      MeilleurCircuit=l; //l*2 techniquement
		      MeilleurSommet=j;
		      Place=a;
		    }
		  a=a->GetNext2();
		}
	      while (a!=Circuit); //on est sur une liste chainee circulaire
	    }
	}//fin boucle pour chaque sommet
      //on insere MeilleurSommet a Place
     SommetsTraites[MeilleurSommet]=1;
     

      //cas particulier Circuit vide
      if (!Circuit)
	{//ATTENTION: ne pas commetre l'erreur de creer une arrete qui boucle sur elle meme
	  tArrete *a1=new tArrete();
	  a1->s1=0;
	  a1->s2=MeilleurSommet;
	  tArrete *a2=new tArrete();
	  a2->s2=0;
	  a2->s1=MeilleurSommet;
	  a1->SetNext2(a2);
	  a1->SetPrev2(a2);
	  a2->SetNext2(a1);
	  a2->SetPrev2(a1);
	  Circuit=a1;
	  circuit->Add(0,MeilleurSommet);
	}
      else
	{
	  if (Output>=Detailed) printf("HMoindreCout: ajout de  %d entre %d et %d\n",MeilleurSommet,Place->s1,Place->s2);
	  //maj circuit
	  if (k!=2) circuit->Remove(Place->s1,Place->s2); //si k=2 cette arete etait en double
	  if (Output>=Debug) printf("circuit rem %d %d\n",Place->s1,Place->s2);	  
	  circuit->Add(MeilleurSommet,Place->s1);
	  if (Output>=Debug) printf("circuit add  %d %d\n",Place->s1,MeilleurSommet);
	  circuit->Add(MeilleurSommet,Place->s2);
	  if (Output>=Debug) printf("circuit rem %d %d\n\n",MeilleurSommet,Place->s2);
	  tArrete *a=new tArrete();
	  a->s1=Place->s1;
	  a->s2=MeilleurSommet;
	  Place->InsertBefore2(a);
	  Place->s1=MeilleurSommet;
	  
	}
      PoidCircuit+=MeilleurCircuit;
   
      if (OPTION_DRAW(Options)) 
	{
	usleep(OPTION_DRAWWAIT(Options)*10000);
	e->Draw(circuit);
	}
    }//fin boucle d'insertion
  if (Output>=Important) printf("HMoindreCout: fin, chemin de longueur %d trouve\n",PoidCircuit);  
  return circuit;
}


tArretes * PVC_HPlusProcheVoisin(EuclidianGraphe *e,unsigned long options)
{
  OutputLevel Output=(OutputLevel)OPTION_OUTPUT(options);
  if (Output>=Important) printf("HPlusProcheVoisin: algorithme du plus proche voisin, graphe de %d sommets\n",e->GetSommets());
  char *SommetsTraites=new char[e->GetSommets()];
  memset(SommetsTraites,0,e->GetSommets());
  tArretes *a=new tArretes(e->GetSommets());
  unsigned int LastSommet; //dernier sommet insere  
  unsigned int MeilleurSommet,MeilleurPoid;  
//initialisation;
  LastSommet=0;//on part du sommet 0
  for (int k=1;k<e->GetSommets();k++)
    {
      if (Output>=UltraDetailed) printf("HPlusProcheVoisin: recherche du %d ieme sommet a inserer\n",k+1);
      MeilleurPoid=0x7fffffff;
      for (int i=1;i<e->GetSommets();i++)
	{
	  if (SommetsTraites[i]) continue;
	  unsigned int l= e->Get(i,LastSommet);
	  if (l<MeilleurPoid)
	  {
	    MeilleurPoid=l;
	    MeilleurSommet=i;
	  }
	}
      //on insere meilleursommet
      if (Output>=Detailed) printf("HPlusProcheVoisin: insertion de %d apres %d\n",MeilleurSommet,LastSommet);
      SommetsTraites[MeilleurSommet]=1;
      a->Add(MeilleurSommet,LastSommet);
      LastSommet=MeilleurSommet;
      if (OPTION_DRAW(options))
	{
	  e->Draw(a);
	  usleep(OPTION_DRAWWAIT(options)*10000);
	}
    }
  //ne pas oublier de boucler le circuit
  a->Add(LastSommet,0);
  delete SommetsTraites;
  if (Output>=Important) printf("HPlusProcheVoisin: fin, circuit de longueur %d trouve\n",e->Eval(a));
  return a;
}




/**********************2-Opt*************************/
tArretes * PVC_H2Opt(EuclidianGraphe *e,unsigned long options)
{
  //note: la principale difficult est de passer d'un circuit a un autre en temps constant
  //en effet si on represente le circuit par une liste chainee, il faut renverser une partie
  //de la liste a chaque etape. La solution consiste a tagger les elements de la liste chainee
  //pour signifier que les pointeurs prev et next sont inverss. On ne tag pas chaque element, mais 
  //chaque point ou il faut inverser le sens theorique de parcour pour garder un sens constant
  //se confrer a la documentation pour une explication plus dtaille
  OutputLevel Output=(OutputLevel)OPTION_OUTPUT(options);
  if (Output>=Important) printf("H2Opt: algorithme de voisinage 2-Opt, dmarrage\n");
  if (Output>=Important) printf("H2Opt: construction d'un chemin initial\n");
  tArretes * ta=PVC_HPlusProcheVoisin(e,options);
  if (Output>=Important) printf("H2Opt: initialisation\n");
  //on mets le chemin sous une forme plus accessible
  tArrete  * chemin=ta->CircuitAsLinkedList();
  if (!chemin)
    {
      if (Output>=Important) printf("H2Opt: erreur, echec de CircuitAsLinkedList\n");
      return NULL;
    }
  //validation du chemin
  tArrete *tmp=chemin;
  for (int i=0;i<e->GetSommets();i++)
    {
      if (tmp->s2!=tmp->GetNext2()->s1) 
	{
	  if (Output>=Important) printf("H2Opt: erreur,chemin invalide\n");
	  return NULL;
	}
      if (Output>=UltraDetailed) printf("H2Opt: %d: %d-%d\n",i,tmp->s1,tmp->s2);
      tmp=tmp->GetNext2();
    }
  if (tmp!=chemin)
    {
      if (Output>=Important) printf("H2Opt: erreur,chemin invalide\n");
      return NULL;
    }
  tArrete * meilleurechange[2]; //{a,b} signifie echange des arretes a et b
  bool reversedstate[2]; //etat des variables reversed pour les meilleures arretes actuelles
  int meilleurgain; //longueur gagnee sur le circuit
  char * reversedsommets=(char *)malloc(e->GetSommets()); 
  memset(reversedsommets,0,e->GetSommets());
  if (Output>=Important) printf("H2Opt: boucle principale\n");
  //!!! ON FAIT EN SORTE QUE L'ARRETE chemin("racine") ne soit jamais inverse( ce qui ne veut pas dire que reversedsommets[0]==0 )
  do
    {
      meilleurgain=0; //on accepte pas les changements negatifs
      //on teste toutes les permutations possibles, soit tous les couples d'arretes non consecutives
      if (Output>=Debug)
	{//code de deboguage, verification de la validite du chemin
	  tArrete *tmp=chemin;
	  unsigned long ts=chemin->s2;
	  bool trev=false;
	  printf("%d: %d-%d\n",0,chemin->s1,chemin->s2);
	  for (int i=1;i<e->GetSommets();i++)
	    {
	      tmp=tmp->GetNext2(trev,reversedsommets);
	      printf("%d: %d-%d(%d)(%d %d)\n",i,tmp->GetS1(trev),tmp->GetS2(trev),trev?1:0,reversedsommets[tmp->GetS1(trev)],reversedsommets[tmp->GetS2(trev)]);
	      if (tmp->GetS1(trev)!=ts) 
		{
		  if (Output>=Important) printf("H2Opt: erreur,chemin invalide( %d eme S1= %d et non %d\n",i,tmp->GetS1(trev),ts);
		  return NULL;
		}
	      ts=tmp->GetS2(trev);
	    }
	  if (tmp->GetNext2(trev)!=chemin)
	    {
	      if (Output>=Important) printf("H2Opt: erreur,chemin invalide\n");
	      return NULL;
	    }
	}//fin code de deboguage
      bool reversedi=false,reversedj=false; 
      //precalcul de la condition d'arret sur i: avant-avant-avant derniere arrete est la derniere testee
      bool temp=false;
      tArrete * istop=chemin->GetPrev2(temp,reversedsommets)->GetPrev2(temp,reversedsommets);
      int loopdetecti=0;
      for (tArrete * i=chemin;i!=istop;i=i->GetNext2(reversedi,reversedsommets))
	{
#ifdef DEBUG
	  loopdetecti++;
	  if (loopdetecti>e->GetSommets())
	    {
	      if (Output>=Important) printf("erreur:boucle dtecte, chemin corrompu\n");
	      return NULL;
	    }
#endif
	  reversedj=reversedi;
	  int loopdetectj=0;
	  for (tArrete * j=i->GetNext2(reversedj,reversedsommets)->GetNext2(reversedj,reversedsommets);j!=chemin;j=j->GetNext2(reversedj,reversedsommets))
	    {
#ifdef DEBUG
	      loopdetectj++;
	      if (loopdetectj>e->GetSommets())
		{
		  if (Output>=Important) printf("erreur:boucle dtecte, chemin corrompu\n");
		  return NULL;
		}
#endif
	      int gain=e->Get(i->s1,i->s2)+e->Get(j->s1,j->s2)
		-e->Get(i->GetS1(reversedi),j->GetS1(reversedj))-e->Get(i->GetS2(reversedi),j->GetS2(reversedj));// inverser s1 et s2 n'a pas d'importance pour les 2 premiers membres, mais en apour les deux suivants
	      if (Output>=UltraDetailed) printf("H2Opt: test de permutation %d-%d  %d-%d , gain de %d\n",i->GetS1(reversedi),i->GetS2(reversedi),j->GetS1(reversedj),j->GetS2(reversedj),gain);
	      if (gain>meilleurgain)
		{
		  meilleurgain=gain;
		  meilleurechange[0]=i;
		  meilleurechange[1]=j;
		  reversedstate[0]=reversedi;
		  reversedstate[1]=reversedj;
		}
	    }
	}
      if (meilleurgain)
	{//on passe au nouveau circuit
	  //l'ordre des operations est important
	  //on inverse le sens de parcours de la moitie i->GetS2-j->GetS1, de sorte que le sens de parcours de l'arrete partant de 0, pointee par chemin, soit toujours non inverse
	  int S2i=meilleurechange[0]->GetS2(reversedstate[0]);
	  int S1i=meilleurechange[0]->GetS1(reversedstate[0]);
	  int S2j=meilleurechange[1]->GetS2(reversedstate[1]);
	  int S1j=meilleurechange[1]->GetS1(reversedstate[1]);
	  if (Output>=Detailed) printf("H2Opt: inversion %d-%d et %d-%d\n",S1i,S2i,S1j,S2j);
	  //note: les sens de parcours des arretes qu'on modifie restent inchanges
	  meilleurechange[0]->SetS2(S1j,reversedstate[0]);
	  meilleurechange[1]->SetS1(S2i,reversedstate[1]);
	  //on modifie les pointeurs
	  tArrete * iNext=meilleurechange[0]->GetNext2(reversedstate[0]);
	  tArrete * jPrev=meilleurechange[1]->GetPrev2(reversedstate[1]);
	  //attention, la modif des pointeurs de inext ne se fait pas avec reversedstate[0]
	  iNext->SetPrev2(meilleurechange[1],( (reversedstate[0]?1:0)+reversedsommets[S2i])%2 );
	  //a noter qu'au final jPrev sera le suivant de iNext, mais le sens de parcour n'est pas encore inverse
	  meilleurechange[1]->SetPrev2(iNext,reversedstate[1]);
	  //les deux lignes precedentes peuvent sembler contradictoire, mais ne pas oublier qu'on va inserer un reverse au sommet reliant meilleurechange[1] et iNext
	  meilleurechange[0]->SetNext2(jPrev,reversedstate[0]);
	  jPrev->SetNext2(meilleurechange[0], ((reversedstate[1]?1:0)+reversedsommets[S1j])%2 );
	  //memes remarques qu'au dessus, situation symtrique
	  //on inverse, cf dessin pour ne pas se tromper de point, i devient S1i-S1j et j devient S2i-S2j 
	  //mise a jour des reversedsommets, il faut regarder S1j et S2i
	  //S2i: on compare reversedstate[0] et reversedstate correspondant a prevj
	  bool revs1j=reversedsommets[S1j];
	  if ( (((reversedstate[0]?1:0)+((reversedstate[1]?1:0)+reversedsommets[S2i]))%2)!=0)
	    {
	      reversedsommets[S2i]=0;
	    }
	  else reversedsommets[S2i]=1;
	  //idem pour reversedstate[1] et le reversedstate de nexti, sans oublier d'inverser car nexti sera parcouru dans l'autre sens (rester calme et faire un dessin)
	  if ( (((reversedstate[1]?1:0)+((reversedstate[0]?1:0)+revs1j))%2)!=0)
	    {
	      reversedsommets[S1j]=0;
	    }
	  else reversedsommets[S1j]=1;
	  //note: on est sur que chemin->s1 est toujours egal a 0, car i->s1 n'a pas boug
	  //on maj le tArretes, qu'on n'utilise plus
	  ta->Remove(S1j,S2j);
	  ta->Remove(S1i,S2i);
	  ta->Add(S1i,S1j);
	  ta->Add(S2i,S2j);
	  if (OPTION_DRAW(options))
	    {
	      e->Draw(ta);
	      usleep(OPTION_DRAWWAIT(options)*10000);
	    }
	  
	}
    }
  while (meilleurgain!=0);
  //on efface la liste chainee circulaire
  bool trev=false;
  tArrete * next=chemin->GetNext2(trev,reversedsommets);
  while (next!=chemin)
  {
    tArrete * todel=next;
    next=next->GetNext2(trev,reversedsommets);
    delete todel;
  }
  delete chemin;
  free(reversedsommets);
 if (Output>=Important) printf("H2Opt: fin, circuit de longueur %d trouve\n",e->Eval(ta    ));
  return ta;
}




/******************************Recuit Simule********************/
tArretes * PVC_HRecuitSimule(EuclidianGraphe *e,unsigned long options)
{
#define TFACTOR 0.93
 OutputLevel Output=(OutputLevel)OPTION_OUTPUT(options);
 unsigned long nbretemp=(OPTION_SPEC(options)&0xf)*5; //nombre de paliers de temperature
 if (!nbretemp) nbretemp=35;
 unsigned long nbretrans=(OPTION_SPEC(options)&0xf0)*e->GetSommets()/4; //nombre de transitions par temperature
 if (!nbretrans) nbretrans=28*e->GetSommets();
 //calcul de la temperature intiiale 
 //e^(-0.7)=0.5 environ
 float tlen=((float)e->GetMaxCoord())/sqrt(e->GetSommets());
 float temperature=(tlen/2.0)/0.7; //une transition ralongeant de tlen/2  a 50% de chances d'etre acceptee initialement
 if (Output>=Important) printf("HRecuitSimule: algorithme du plus proche voisin, graphe de %d sommets, %d paliers de %d transitions, temperature initiale %f\n",e->GetSommets(),nbretemp,nbretrans,temperature);
 if (Output>=Important) printf("HRecuitSimule: recherche d'une solution initiale\n");
 tArretes *ta=PVC_HPlusProcheVoisin(e,options);
 if (!ta) return NULL;
 //mise en place des structures necessaires aux transitions 2opt
 tArrete  * chemin=ta->CircuitAsLinkedList();
 if (!chemin)
   {
     if (Output>=Important) printf("H2Opt: erreur, echec de CircuitAsLinkedList\n");
     return NULL;
   }
 char * reversedsommets=(char *)malloc(e->GetSommets()); 
 memset(reversedsommets,0,e->GetSommets());
 //afin d'augmenter le cot alatoire, on ne part pas du sommet 0apres chaque transition, mais d'n sommet tir au hasard( on stoque la liste des sommets dans un tableau pour eviter tout parcours inutile de la liste chainee
 tArrete **depart=(tArrete **)malloc(sizeof(tArrete *)*e->GetSommets());
 tArrete *ar=chemin;
 for (int i=0;i<e->GetSommets();i++)
   {
     depart[i]=ar;
     ar=ar->GetNext2();
   }
 tArrete * meilleurechange[2];
 bool reversedstate[2];
 bool transitionok;
 for (int bi=0;bi<nbretemp;bi++,temperature=temperature*TFACTOR)
   {//boucle temperature
     if (Output>=Detailed) printf("HRecuitSimule: temperature=%f, (circuit actuel: %d)\n",temperature,e->Eval(ta));
     for (int bj=0;bj<nbretrans;bj++)
       {//boucle transition
	 if (Output>=Verbose) printf("transition %d sur %d\n",bj,nbretrans);
	 transitionok=false;
	 //choix du sommet de depart
	 tArrete * dep=depart[rand()%e->GetSommets()];
	 bool reversedi=false,reversedj=false; 
	 //precalcul de la condition d'arret sur i: avant-avant-avant derniere arrete est la derniere testee
	 bool temp=false;
	 tArrete * istop=dep->GetPrev2(temp,reversedsommets)->GetPrev2(temp,reversedsommets);
	 for (tArrete * i=dep;i!=istop;i=i->GetNext2(reversedi,reversedsommets))
	   {
	     reversedj=reversedi;
	     for (tArrete * j=i->GetNext2(reversedj,reversedsommets)->GetNext2(reversedj,reversedsommets);j!=chemin;j=j->GetNext2(reversedj,reversedsommets))
	       {
		 
		 int gain=e->Get(i->s1,i->s2)+e->Get(j->s1,j->s2)
		   -e->Get(i->GetS1(reversedi),j->GetS1(reversedj))-e->Get(i->GetS2(reversedi),j->GetS2(reversedj));
		 if (gain>0)
		   {//acceptation automatique
		     if (Output>=Verbose) printf("gain de %d, acceptation automatique\n");
		     transitionok=true;
		     meilleurechange[0]=i;
		     meilleurechange[1]=j;
		     reversedstate[0]=reversedi;
		     reversedstate[1]=reversedj;
		     break;
		   }
		 //sinon, test
		 float seuil=exp(gain/temperature)*RAND_MAX;
		 int rnd=rand();
		 if (Output>=Verbose) printf("gain de %d, probabilite de transition de %f/%d, jet de %d\n",gain,seuil,(int)RAND_MAX,rnd);
		 if (seuil>rnd)
		   {
		     transitionok=true;
		     meilleurechange[0]=i;
		     meilleurechange[1]=j;
		     reversedstate[0]=reversedi;
		     reversedstate[1]=reversedj;
		     break;
		   }
	       }
	     if (transitionok)
	       {//code recopie de H2Opt
		 int S2i=meilleurechange[0]->GetS2(reversedstate[0]);
		 int S1i=meilleurechange[0]->GetS1(reversedstate[0]);
		 int S2j=meilleurechange[1]->GetS2(reversedstate[1]);
		 int S1j=meilleurechange[1]->GetS1(reversedstate[1]);
		 if (Output>=UltraDetailed) printf("HRecuitSimule: inversion %d-%d et %d-%d\n",S1i,S2i,S1j,S2j);
		 meilleurechange[0]->SetS2(S1j,reversedstate[0]);
		 meilleurechange[1]->SetS1(S2i,reversedstate[1]);
		 tArrete * iNext=meilleurechange[0]->GetNext2(reversedstate[0]);
		 tArrete * jPrev=meilleurechange[1]->GetPrev2(reversedstate[1]);
		 iNext->SetPrev2(meilleurechange[1],( (reversedstate[0]?1:0)+reversedsommets[S2i])%2 );
		 meilleurechange[1]->SetPrev2(iNext,reversedstate[1]);
		 meilleurechange[0]->SetNext2(jPrev,reversedstate[0]);
		 jPrev->SetNext2(meilleurechange[0], ((reversedstate[1]?1:0)+reversedsommets[S1j])%2 );
		 bool revs1j=reversedsommets[S1j];
		 if ( (((reversedstate[0]?1:0)+((reversedstate[1]?1:0)+reversedsommets[S2i]))%2)!=0)
		   {
		     reversedsommets[S2i]=0;
		   }
		 else reversedsommets[S2i]=1;
		 if ( (((reversedstate[1]?1:0)+((reversedstate[0]?1:0)+revs1j))%2)!=0)
		   {
		     reversedsommets[S1j]=0;
		   }
		 else reversedsommets[S1j]=1;
		 ta->Remove(S1j,S2j);
		 ta->Remove(S1i,S2i);
		 ta->Add(S1i,S1j);
		 ta->Add(S2i,S2j);
		 if (OPTION_DRAW(options))
		   {
		     e->Draw(ta);
		     usleep(OPTION_DRAWWAIT(options)*10000);
		   }
		 break;
	       }   
	   }
	 
       }
   }
 bool trev=false;
 tArrete * next=chemin->GetNext2(trev,reversedsommets);
 while (next!=chemin)
   {
     tArrete * todel=next;
     next=next->GetNext2(trev,reversedsommets);
     delete todel;
   }
 delete chemin;
 free(reversedsommets);
 free(depart);
 if (Output>=Important) printf("HRecuitSimule: fin, circuit de longueur %d trouve\n",e->Eval(ta    ));
 return ta;
}




//***********************************Separation/Evaluation**************************/
class Arbre:public xTree
{
public:
  unsigned long s;//le sommet correspondant a ce noeud
  unsigned long borneinf;
  bool ferme; //true si cette branche a t abandonnee
};

//cette fonction efface un arbre, de maniere non recursive
void ClearArbre(Arbre * root)
{
  while ((root)&&((root->Parent)||(root->Child)||(root->Next)||(root->Prev)))
    {
      if (root->Child) {root=(Arbre *)root->Child;continue;}
      if (root->Next) {root=(Arbre *)root->Next;continue;}
      Arbre *pr=(Arbre *)root->Prev,*pa=(Arbre *)root->Parent;
      delete root;
      root=pr?pr:pa;
    }
}

char * ASEBuff=NULL; //buffer global, allocations multiples=temps perdu
OutputLevel ASEOutputLevel;
unsigned long ASEOptions;
tArretes * ASEArretes; //utilises seulement si ASEOptions inclu l'affichage
//fonction d'evaluation, renvoit la borne sup d'une solution sur cette branche et si elle est atteinte
unsigned long ASEEval(EuclidianGraphe *e,int &atteinte,char * usedsommets,int CurrentSommet,unsigned long * chemin, int longueurchemin,bool continuesibornenonatteinte)
{
  //on fait un melange iteratif/recursif, selon qu'il y a un cas ou plusieurs a traiter
  //en effet le cas ou plusieurs arretes ont le meme poid minimal est rare, et la surcharge aporte par la rcursivite est lourde  
  //atteint: (sortie) non nul si la borne est atteinte
  //usedsommets: sommets deja atteints
  //longueurchemin: nombre de sommets dans le chemin actuel
  //continuesobornenonatteinte: false<=> si borne non atteinte, ne pas calculer la borne
  unsigned long currentp=0; //poid de la partie de chemin generee
  if (ASEOutputLevel>=Verbose) printf("+\n");
  while (true)
    {
      if (ASEOutputLevel>=Verbose) printf("+%d, %d\n",CurrentSommet,currentp); 
      if (longueurchemin==e->GetSommets()) //usedsommets vaut 1 partout sauf en CurrentSommet
	{
	  //il reste a boucler: il faut que min w(i,CurreentSommet) soit atteint pour i=chemin[0]
	  unsigned long minp=-1;
	  for (int i=0;i<e->GetSommets();i++)
	    {
	      if (i==CurrentSommet) continue;
	      if (e->Get(i,CurrentSommet)<minp) minp=e->Get(i,CurrentSommet);
	    }
	  atteinte=(minp==e->Get(CurrentSommet,chemin[0]));
	  return currentp+minp;
	}
      unsigned long minp=-1;//poid min
      unsigned long count=1;//nbre de sommets realisant le min
      unsigned long mins=-1; //sommet realisant le min
      //on cherche min{ w(i,currentsommet}, ainsi que le nombre d'arretes le 
      //realisant, telles que i ne soit pas deja atteint
      for (int i=0;i<e->GetSommets();i++)
	{
	  if (i==CurrentSommet) continue;
	  if ((e->Get(i,CurrentSommet)==minp)&&(!usedsommets[i])) {count++;mins=i;}
	  if (e->Get(i,CurrentSommet)<minp) 
	    {
	      minp=e->Get(i,CurrentSommet);
	      count=0;
	      mins=-1;
	      if (!usedsommets[i])
		{
		  count=1;
		  mins=i;
		}
	    }
	}
      if (count==0)
	{//on est deja passe par le sommet qui realise le min:donc  borne non atteinte
	  atteinte=0; 
	  if (!continuesibornenonatteinte) return -1;
	  currentp+=minp;
	  usedsommets[CurrentSommet]=1;
	  //on sait deja que la borne est non atteinte: on procede diffremment
	  //on enumere les sommets non encore traites
	  for (int i=0;i<e->GetSommets();i++)
	    {
	      if (usedsommets[i]) continue;
	      //on calcule le min pour ce sommet
	      unsigned long minp=-1;
	      for (int j=0;j<e->GetSommets();j++)
		{
		  if (i==j) continue;
		  if (e->Get(i,j)<minp) minp=e->Get(i,j);
		}
	      //on rajoute a la somme
	      currentp+=minp;
	      if (ASEOutputLevel>=Debug) printf("+>%d (%d)\n",i,currentp);
	    }
	  //et on renvoi le resultata
	  return currentp;
	}
      //si le min est unique, on itere
      if (count==1)
	{
	  currentp+=minp;
	  usedsommets[CurrentSommet]=1;
	  chemin[longueurchemin++]=mins;
	  CurrentSommet=mins;
	  continue;
	}
      else
	{//sinon on s'apelle recursivement pour chacune des possibilites
	  //on sait que chaque possibilite generera la meme longueur
	  //ATTENTION: il est inutile de dupliquer chemin, mais il faut copier usedsommets
	  char * _usedsommets=(char *)malloc(e->GetSommets());
	  currentp+=minp;
	  usedsommets[CurrentSommet]=1;
	  int traite=0; //pour accelerer un peu
	  int _atteinte=1;
	  unsigned long longueur=-1;
	  for (int i=0;i<e->GetSommets()&&(traite<count);i++)
	  {
	    if ((!usedsommets[i])&&(e->Get(i,CurrentSommet)==minp))
	      {
		traite++;
		chemin[longueurchemin]=i;
		_atteinte=1;
		memcpy(_usedsommets,usedsommets,e->GetSommets());
		if (longueur==-1) longueur=ASEEval(e,_atteinte,_usedsommets,i,chemin,longueurchemin+1,true);
		else ASEEval(e,_atteinte,_usedsommets,i,chemin,longueurchemin+1,false);
		if (_atteinte==1) 
		  {
		    atteinte=1;
		    free(_usedsommets);
		    return longueur+currentp;
		  }
	      }
	  }
	  free(_usedsommets);
	  atteinte=0;
	  return longueur+currentp;
	}
    } 
}
 
//fonction recursive de parcours de l'arbre
//renvoit 0 si plus aucune expanssion possible
unsigned long ASEParcours(Arbre * a,EuclidianGraphe *e,unsigned long &bestsolution,unsigned long * chemin,unsigned long position, char * usedsommets,unsigned long * meilleurchemin,unsigned long currentlength)
{
  //a: noeud actuel de l'arbre
  //currentlength= longueur de chemin[0]...chemin[?]-a->s
  //meilleur chemin: meilleur chemin actuel, qu'on remplace si on trouve mieux
  //bestsolution: longueur du meilleur chemin
  //chemin[0]..chemin[position-1]: le sous chemin qui nous sert de racine
  //usedsommets: usedsommets[i]!=0 <=> le sous chemin racine passe par i
  if (!a) return 0;  
//parcours l'arbre, developpe un niveau
  chemin[position]=a->s;
  if (ASEOutputLevel>=UltraDetailed) printf("=>%d   (borne %d, sous chemin %d)\n",a->s,a->borneinf,currentlength);
  if (a->ferme) 
    {
      if (ASEOutputLevel>=UltraDetailed) printf("ferme!\n<=%d\n",a->s);
      return 0;
    }
  if (a->borneinf>bestsolution)
    {
      if (ASEOutputLevel>=UltraDetailed) printf("fermeture: borne inf %d > meilleure solution %d!\n",a->borneinf,bestsolution);
      a->ferme=true;
      //on efface tous les descendants
      if (a->Child) Eradicate<Arbre>((Arbre *)a->Child);
      a->Child=NULL; //si on ne le fait pas, crash quand on liberera la memoire
      return 0;
    }
  if (a->Child)
    {//nous sommes un noeud intermediaire
      usedsommets[a->s]=1;
      unsigned long ret=ASEParcours((Arbre *)a->Child,e,bestsolution,chemin,position+1,usedsommets,meilleurchemin,currentlength+e->Get(a->s,((Arbre *)a->Child)->s));
      usedsommets[a->s]=0;
      if (ASEOutputLevel>=UltraDetailed) printf("<=%d\n",a->s);
      if (a->Next)
	{
	  //attention au calcul de currentlength, 
	  //on enleve de a->Parent a a, on rajoute de a->Parent a a->Next
	  unsigned long deltalen=e->Get( ((Arbre *)a->Parent)->s,((Arbre *)a->Next)->s) 
	    - e->Get( ((Arbre *)a->Parent)->s,a->s);
	  unsigned long ret2=ASEParcours(
			 (Arbre *)a->Next,e,bestsolution,chemin,position,
			 usedsommets,meilleurchemin,
			 currentlength+deltalen
			 );
	  return (ret||ret2);
	}
      else return ret;
    }
  else
    {//nous sommes un sommet terminal de l'arbre: on etend d'un  niveau
      //note:il est inutile d'inserer un sommet qui sera immediatement ferme
      if (OPTION_DRAW(ASEOptions))
	{
	  //inclure une gestion efficace de l'affichage alourdirait encore ce code
	  //et nuirait grandement a sa lisibilit
	  //etant donn que si la vitesse est un critre important, l'affichage sera coup
	  //ce n'est pas rellement un problme
	  ASEArretes->Clear();
	  Arbre * ta=a;
	  while (ta->Parent)
	    {
	      ASEArretes->Add(ta->s,((Arbre *)ta->Parent)->s);
	      ta=(Arbre *)ta->Parent;
	    }
	  e->Draw(ASEArretes);
	  usleep(OPTION_DRAWWAIT(ASEOptions)*10000);
	} 
      //si on est a la profondeur max, on examine le chemin
      if (position==e->GetSommets()-1)
	{
	  unsigned long clen=currentlength+e->Get(chemin[0],a->s);
	  if (clen<bestsolution)
	    {
	      bestsolution=clen;
	      for (int i=0;i<e->GetSommets();i++)
		meilleurchemin[i]=chemin[i];
	      if (ASEOutputLevel>=UltraDetailed) printf("ASepEval: Nouvelle meilleure solution:%d\n",bestsolution);
	    }
	  a->ferme=true;
	}
      else
	{//sinon: on developpe ses fils
	  if (ASEOutputLevel>=UltraDetailed) printf("expand:\n");
	  usedsommets[a->s]=1;
	  for (int i=0;i<e->GetSommets();i++)
	    {
	      if ((usedsommets[i])) continue;
	      chemin[position+1]=i;
	      int  atteint=1;
	      memcpy(ASEBuff,usedsommets,e->GetSommets());
	      unsigned long length=currentlength+e->Get(a->s,i)+ASEEval(e,atteint,ASEBuff,i,chemin,position+2,true);
	      if (ASEOutputLevel>=UltraDetailed) printf("->%d borne inf %d (%i+%i)(%s)\n",i,length,currentlength,length-currentlength,atteint?"atteint":"non atteint");
	      if (length>bestsolution) continue; //on ignore
	      Arbre * ta=new Arbre();
	      ta->ferme=false;
	      ta->s=i;
	      ta->borneinf=length;
	      a->AddChild(ta);
	      if ((atteint)&&(length<bestsolution))
		{
		  bestsolution=length;
		  memcpy(meilleurchemin,chemin,sizeof(unsigned long)*e->GetSommets());
		}
	    }
	  if (!a->Child)
	    {//tous les fils ont ete elagues
	      a->ferme=true;
	    }
	  usedsommets[a->s]=0;
	}
      if (ASEOutputLevel>=UltraDetailed) printf("<=%d\n",a->s);
      if (a->Next)
	{
	  unsigned long deltalen=e->Get( ((Arbre *)a->Parent)->s,((Arbre *)a->Next)->s) 
	    - e->Get( ((Arbre *)a->Parent)->s,a->s);
	  if (!a->Child)
	    return ASEParcours((Arbre *)a->Next,e,bestsolution,chemin,position,usedsommets,meilleurchemin,currentlength+deltalen);
	  else
	    {
	      ASEParcours((Arbre *)a->Next,e,bestsolution,chemin,position,usedsommets,meilleurchemin,currentlength+deltalen);
	      return 1;
	    }
	}
      //pas de frere a droite
      else return a->Child?1:0;
    }
}

tArretes * PVC_ASeparationEvaluation(EuclidianGraphe *e, unsigned long options)
{
 OutputLevel Output=(OutputLevel)OPTION_OUTPUT(options);
 ASEOutputLevel=Output;
 ASEOptions=options;
 if (OPTION_DRAW(options)) ASEArretes=new tArretes(e->GetSommets());
 if (Output>=Important) printf("ASepEval: algorithme exact de sparation/evaluation, dmarrage\n");
 ASEBuff=(char *)malloc(e->GetSommets());
 char * usedsommets=(char *)malloc(e->GetSommets());
 memset(usedsommets,0,e->GetSommets());
 unsigned long * meilleurchemin; //=(unsigned long *)malloc(sizeof(unsigned long)*e->GetSommets());
 unsigned long * chemin=(unsigned long *)malloc(sizeof(unsigned long)*e->GetSommets());
 unsigned long bestlength=-1;
 Arbre * root=new Arbre();
 root->s=0;
 root->ferme=false;
 int  inu;
 //afin d'ameliorer l'elagation, on part d'un chemin pas trop mauvais gnr par un algorithme pas trop lent
 //2Opt  par exemple
 tArretes *ta=PVC_H2Opt(e,options);
 bestlength=e->Eval(ta);
 meilleurchemin=(unsigned long *)ta->CircuitAsArray();
 delete ta;
 chemin[0]=0;
 root->borneinf=ASEEval(e,inu,usedsommets,0,chemin,1,true);
 memset(usedsommets,0,e->GetSommets());
 //on apelle la fonction de parcours tant qu'elle peut continuer
 unsigned long prof=0;
while (ASEParcours(root,e,bestlength,chemin,0,usedsommets,meilleurchemin,0))
   {
     prof++;
     if (Output>=Detailed)
       {
	 printf("ASepEval: profondeur %d, meilleure solution: %d\n",prof,bestlength);
       }
   //debug
     /*
   for (int i=0;i<e->GetSommets();i++)
     {
       if (usedsommets[i]) printf ("BUG\n");
     }
     */
   // /debug
   }
 if (Output>=Important) printf("ASepEval: fin\n");
 //on converti le meilleur chemin en tArretes
 if (Output>=UltraDetailed) printf("ASepEval: solution(%d)\n",bestlength);
 ta=new tArretes(e->GetSommets());
 for (int i=0;i<e->GetSommets()-1;i++)
   {
   ta->Add(meilleurchemin[i],meilleurchemin[i+1]);
   if (Output>=UltraDetailed) printf("ASepEval: %d-%d\n",meilleurchemin[i],meilleurchemin[i+1]);
   }
 ta->Add(meilleurchemin[0],meilleurchemin[e->GetSommets()-1]);
 //on nettoie 

 //le probleme de la fonction Eradicate de xtree.hpp est qu'elle est rcursive:
 //dans le cas d'arbres de taille importante, la place occupe sur la pile devient exorbitante
 //Eradicate<Arbre>((Arbre *)root->Child);
 ClearArbre(root);
 free(meilleurchemin);
 free(chemin); 
 free(ASEBuff); 
 free(usedsommets);
 if (OPTION_DRAW(options)) delete ASEArretes;
 if (OPTION_DRAW(options))
   {
     e->Draw(ta);
     usleep(OPTION_DRAWWAIT(options)*10000);
   } 
 if (Output>=Important) printf("ASepEval: chemin optimal de longueur %d trouv\n",e->Eval(ta));
 return ta;
}





/**************************ELASTIQUE****************************/
  //j'opte pour une implmentation multithread, supportant donc le multiprocessing, car ces 
  //un des principaux interets de cet algorithme: il est massivement paralllisable
 
struct fPOINT
{
  double  x;
  double y;
};

#define HEL_MAX_THREADS 4  //nbre max de threads a lancer, valeur ideal= le nombre de processeurs disponibles

//la structure suivante definit un sommet de l'elastique
//elle contient toutes les informations necessaires a un thread pour traiter ce sommet a chaque iteration
class SommetInfo
{
public:
  SommetInfo * NextSommet; //!! sommet suivant a traite par ce thread
  unsigned long Id; //index dans les tableaux de ce sommet
  unsigned long NextId;//index du sommet suivant dans l'elastique
  unsigned long PrevId;//index du sommet precedent dans l'elastique
  fPOINT NewPos; //variable ou stoquer temporairement la position calculee
  double DNext; //distance a NextId
  unsigned long HandleSum[6];  //thread responsable du calcul de ces indices de WeightSum, tableau termine par -1
};
enum HELSTATE
  {
    Weight=0, //calcul des Wij non normalises
    WeightSum, //calcul des sommes sur les villes (HELWeightSum)
    Normalize, //normalisation
    Coords,     //calcul des nouvelles coordonnes des sommets
    Apply,       //mise a jour des coordonnes (on travaille de manire synchrone)
    Die          //fin: terminer le thread
  };
fPOINT * HELCoords; //coordonnees des somets de l'elastique
double **HELWeights; //les Wij i=indicd de ville, j=indice de sommet de l'elastique
double * HELWeightSum; //WeightSum[i]=sum(j=0..M,Wij) : coeff de normalisation
double  HELAlpha;
double HELBeta;
double HELK;
unsigned long HELNumSommets; //nombre de sommets dans l'elastique
HELSTATE HELState;   //variable indiquant aux threads l'action a executer
synchronizer * HELSem; //pour la synchronisation
EuclidianGraphe * HELe;
OutputLevel HELOutput;
SommetInfo *ElSommets; //le tableau de SommetInfo

//cette fonction dessine l'elastique
void HEL_DrawElastique()
{
  /*  glBegin(GL_LINE_STRIP);
  glColor3f(1,1,1);
  int csom=0;
  do
    {
      glVertex2f(HELCoords[csom].x,HELCoords[csom].y);
      csom=ElSommets[csom].NextId;
    }
  while (csom!=0);
  glVertex2f(HELCoords[0].x,HELCoords[0].y);
  glEnd();
  */
}

//code des threads 'esclaves'

void * HEL_WorkingThread(void * p)
{
  SommetInfo *s=(SommetInfo *)p; //sommet racine pour ce thread, constant pour tout le processus
  while(1)
    {
      synchronizer_waitforsignal(HELSem); //attente du signal du thread 'master'
      if (HELOutput>=Verbose) printf("thread %d: got signal for op %d\n",s->Id,(int)HELState);
      if (HELState==Weight)
	{//calcul des Wij pour tout i, pour tout j a notre charge
	  double normfactor=HELe->GetMaxCoord()*HELe->GetMaxCoord();
	  double moinsdeuxKcarre=2.0*HELK*HELK*(-1);
	  for (SommetInfo *t=s;t;t=t->NextSommet)
	    {
	      int j=t->Id;
	      for (int i=0;i<HELe->GetSommets();i++)
		{
		  Point2D * c=HELe->GetCoords(i);
		  double dist= 
		    (( ((double)c->x) - HELCoords[j].x ) * ( ((double)c->x)- HELCoords[j].x))
		    + (( ((double)c->y)- HELCoords[j].y ) * ( ((double)c->y)- HELCoords[j].y));
		  dist/=normfactor;
		  HELWeights[i][t->Id]=exp(dist/moinsdeuxKcarre);
		  if (HELOutput>=Verbose) printf("HELWeight[%d][%d]=%f =exp(%f/%f)\n",i,t->Id, HELWeights[i][t->Id],dist,moinsdeuxKcarre);
		}
	    }	    
	  continue;
	}
      if (HELState==WeightSum)
	{
	  for (SommetInfo *t=s;t;t=t->NextSommet)
	    {
	      for (int i=0;(i<6)&&(t->HandleSum[i]!=-1);i++)
		{
		  int cty=t->HandleSum[i];
		  double sum=0;
		  for (int j=0;j<HELNumSommets;j++) sum+=HELWeights[cty][j];
		  HELWeightSum[cty]=sum;
		}
	    }
	  continue;
	}
      if (HELState==Normalize)
	{
	  for (SommetInfo *t=s;t;t=t->NextSommet)
	    {
	      for (int i=0;i<HELe->GetSommets();i++)
		{
		  HELWeights[i][t->Id]/=HELWeightSum[i];
		}
	    }
	  continue;
	}
      if (HELState==Coords)
	{
	  for (SommetInfo *t=s;t;t=t->NextSommet)
	    {//on applique la grosse formule
	      double X,Y;
	      for (int i=0;i<HELe->GetSommets();i++)
		{
		  Point2D *c=HELe->GetCoords(i);
		  X+=(c->x-HELCoords[t->Id].x)*HELWeights[i][t->Id];
		  Y+=(c->y-HELCoords[t->Id].y)*HELWeights[i][t->Id];
		}
	      X*=HELAlpha;
	      Y*=HELAlpha;
	      X+=(HELCoords[t->NextId].x+HELCoords[t->PrevId].x-2*HELCoords[t->Id].x)*HELBeta*HELK; 
	      Y+=(HELCoords[t->NextId].y+HELCoords[t->PrevId].y-2*HELCoords[t->Id].y)*HELBeta*HELK;
	      t->NewPos.x=HELCoords[t->Id].x+X;
	      t->NewPos.y=HELCoords[t->Id].y+Y;
	      if (HELOutput>=Verbose) printf("thread %d: sommet %d, position %f,%f se deplace de %f,%f\n",s->Id,t->Id,HELCoords[t->Id].x,HELCoords[t->Id].y,X,Y);
	    }
	  continue;
	}
      if (HELState==Apply)
	{
	   for (SommetInfo *t=s;t;t=t->NextSommet)
	     {
	       HELCoords[t->Id].x=t->NewPos.x;
	       HELCoords[t->Id].y=t->NewPos.y;
	       int i=t->Id;
	       int j=t->NextId;
	       double dist= 
		 (( HELCoords[i].x- HELCoords[j].x) *  ( HELCoords[i].x- HELCoords[j].x))
	  +  (( HELCoords[i].y- HELCoords[j].y) *  ( HELCoords[i].y- HELCoords[j].y));
	       t->DNext=sqrt(dist);
	       if (HELOutput>=Verbose) printf("thread %d: sommet %d: distance au suivant %f\n",s->Id,t->Id,t->DNext);
	     }
	   continue;
	}
      if (HELState==Die)
	{
	  if (HELOutput>=Verbose) printf("thread %d: die and release\n",s->Id);
	  pthread_exit(NULL);
	}
    }
}

//fonction principale, thread controleur
tArretes * PVC_HElastique(EuclidianGraphe *e, unsigned long options)
{
  OutputLevel Output=(OutputLevel)OPTION_OUTPUT(options);
  HELOutput=Output;
  if (Output>=Important) printf("HElastique: heuristique de l'elastic net, demarrage\n");
 unsigned long ElIterations=(OPTION_SPEC(options)&0xf)*4; //nombre d'iterations par val de K
 if (!ElIterations) ElIterations=25;
 double ElKFactor=((double)(OPTION_SPEC(options)&0xf0))*0.1+0.84; 
 if (!OPTION_SPEC(options)&0xf0) ElKFactor=0.99;
  HELe=e;
  unsigned long ElMaxSommets=(e->GetSommets()*5)/2; //nombre max de sommets dans l'elastique
  HELNumSommets=(e->GetSommets()*3)/2;
  unsigned long ElMaxThreads=HEL_MAX_THREADS; //cette valeur peut etre quelconque, idalement c'est le nombre de processeurs disponibles
  unsigned long ElThreads=0;
  unsigned long * ElThreadFirstSommet; //indice dans ElSommets du premier sommet gere par chaque thread
  if (e->GetSommets()%5)HELNumSommets++; //necessaire pour le HandleSum
  if (HELNumSommets<3)HELNumSommets=3;
  //allocation des structures de donnee
  HELSem=synchronizer_new();
  ElThreadFirstSommet=(unsigned long  *)malloc(sizeof(unsigned long)*ElMaxThreads);
  HELCoords=(fPOINT *)malloc(sizeof(fPOINT)*ElMaxSommets);
  HELWeights=(double **)malloc(sizeof(double *)*e->GetSommets());
  for( int i=0;i<e->GetSommets();i++)
    {
      HELWeights[i]=(double *)malloc(sizeof(double)*ElMaxSommets);
    }
  HELWeightSum=(double *)malloc(sizeof(double)*e->GetSommets());
  ElSommets=(SommetInfo *)malloc(sizeof(SommetInfo)*ElMaxSommets);
  //initialisation des structures
  //on part de quelques sommets ( 2.5  fois moins que la valeur max) disposes en cercle
  //on effectuera une regle de trois pour ajouter des sommets, sur la base
  //nbre max de sommets pour longueur elastique= e->GetRange()*PI/2
  HELAlpha=0.2;
  HELBeta=2.0;
  HELK=0.2;
  double rayon=0.5*((double)e->GetMaxCoord())/5.0;
  ElThreads=HEL_MAX_THREADS;
  if (ElThreads>HELNumSommets) ElThreads=HELNumSommets;
  //il faut attribuer a chaque thread un certain nombre de sommets
  //et a chaque sommet un certain nombre de HandleSum (villes pour lesquelles
  //il s'occupera de calculer la somme des Wij)
  //cette attribution evoluera au fur et a mesure que des sommets seront 
  //rajoutes a l'elastique
  //on procede ainsi: si on a t thread et s=t*n+r sommets, on attribue 
  //n+1 sommets aux r premiers threads, et n aux restants
  //(les threads etant ordonns selon ElThreadFirstSommet
  //ca permet d'avoir un algorithme de rerepartition simple (cf plus bas)
  //on procede de meme pour les handlesum, les sommets etant ordonnes selon ElSommets
  unsigned long ElFirstFavThread; //premier thread favorise (=r)
  unsigned long ElFirstFavSommet;
  unsigned long ElSPerT; //Sommets par thread (=n)
  unsigned long ElHPerS; //HandleSum par sommet
  unsigned long csommet=0; //nbre de sommets alloues a des threads
  ElSPerT=HELNumSommets/ElThreads;
  ElFirstFavThread=HELNumSommets%ElThreads;
  ElHPerS=e->GetSommets()/HELNumSommets;
  ElFirstFavSommet=e->GetSommets()%HELNumSommets;
  unsigned long currentcity=0; //pour les handlesum
  for (int i=0;i<ElThreads;i++)
    {
      if (csommet==HELNumSommets)
	{//on a fini prematurement
	  ElThreads=i;
	  break;
	}
      ElThreadFirstSommet[i]=csommet;
      int  s=ElSPerT; 
      if (i<ElFirstFavThread) s++; //ces threads la en ont un de plus 
      for (int j=0;j<s;j++)
	{ //traitement du sommet j+csommet
	  //on dispose les sommets de l'elastique en cercle
	  HELCoords[j+csommet].x=((double)e->GetMaxCoord())/2+rayon*cos(3.1415926535*2.0*(j+csommet)/((double)HELNumSommets));
	  HELCoords[j+csommet].y=((double)e->GetMaxCoord())/2+rayon*sin(3.1415926535*2.0*(j+csommet)/((double)HELNumSommets));
	  if (Output>=UltraDetailed) printf("HElastique: positionnement du sommet %d en %f,%f\n",j+csommet, HELCoords[j+csommet].x, HELCoords[j+csommet].y);
	  SommetInfo *si=(SommetInfo *) &ElSommets[j+csommet];
	  si->Id=j+csommet; 
	  si->NextId=j+csommet+1;
	  if (si->NextId>=HELNumSommets) si->NextId=0;
	  si->PrevId=j+csommet-1;
	  if (si->PrevId==(unsigned long)-1) si->PrevId=HELNumSommets-1;
	  si->NextSommet= (j==s-1)? NULL:( &ElSommets[j+csommet+1]);
	  for (int k=0;k<ElHPerS+1;k++)
	    {
	      if ((j+csommet>=ElFirstFavSommet)&&(k==ElHPerS))
		{
		  si->HandleSum[k]=-1;
		}
	      else si->HandleSum[k]=currentcity++;
	    }
	  si->HandleSum[ElHPerS+1]=-1;
	}
      csommet+=s;
      //creation du thread
      pthread_t inu;
      pthread_create(&inu,NULL,HEL_WorkingThread,&ElSommets[ElThreadFirstSommet[i]]);
      //!! on passe en parametre le premier sommet: il ne doit PAS CHANGER
    }



  // on boucle, nous sommes le thread controleur
  synchronizer_waitforwaiters(HELSem,ElThreads); //attente que les threads soient pret
  while (HELK>=0.015)
    {
      if (Output>=Detailed) printf("HElastique: K=%f\n",HELK);
      for (int it=0;it<ElIterations;it++)
	{
	  if (Output>=UltraDetailed) printf("HElastique: iteration...\n");
	  if (OPTION_DRAW(options))
	    {
	      e->Draw(NULL,HEL_DrawElastique);
	      usleep(OPTION_DRAWWAIT(options)*10000);
	    }
	  //une passe
	  for (HELState=Weight;HELState<=Apply;HELState=(HELSTATE)(((int)HELState)+1))
	    {
	      if (Output>=UltraDetailed) printf("HElastique: phase %d\n",(int)HELState);
	      //debloquage 
	      synchronizer_signal(HELSem);
	      //attente de fin d'operation
	      synchronizer_waitforwaiters(HELSem,ElThreads);
	    }
	  //passe terminee

	}
      //serie d'iterations termines
      //on decremente K
      HELK=HELK*ElKFactor;
      //on regarde s'il convient de rajouter des sommets a l'elastique
      //pour cela on calcule sa longueur
      double longueur=0;
      for (int k=0;k<HELNumSommets;k++) longueur+=ElSommets[k].DNext;
      if (Output>=Detailed) printf("HElastique: longueur=%f\n",longueur);
      int stoadd=((double)ElMaxSommets)*longueur/(((double)e->GetMaxCoord())*3.141592/1.5)-HELNumSommets;
      if (stoadd+HELNumSommets>ElMaxSommets) stoadd=ElMaxSommets-HELNumSommets;
      if (stoadd>e->GetSommets()/10+1)  //pour 100 villes on rajoute les sommets par paquet de 10 au moins
	{
	  if (Output>=UltraDetailed) printf("HElastique: ajout de %d sommets\n",stoadd);
	  //on commence par creer et positionner les sommets dans l'elastique
	  //on s'occupera plus tard de les attribuer a des threads
	  //on a deux choix possible: equirepartir entre les sommets existants 
	  //ou sur la longueur , soit inserer un sommet tous les n sommets ou tous
	  //les m metres 
	  //on choisit la premiere solution
	  int insertevry=HELNumSommets/stoadd;
	  int currentsommet=0;
	  for( int i=0;i<stoadd;i++)
	    {
	      //on insere apres currentsommet
	      ElSommets[HELNumSommets+i].Id=HELNumSommets+i;
	      ElSommets[HELNumSommets+i].NextId=ElSommets[currentsommet].NextId;
	      ElSommets[currentsommet].NextId=HELNumSommets+i;
	      ElSommets[HELNumSommets+i].PrevId=currentsommet;
	      //on positionne entre les deux
	      int oldnextid= ElSommets[HELNumSommets+i].NextId;
	      HELCoords[HELNumSommets+i].x=0.5*(HELCoords[oldnextid].x+HELCoords[currentsommet].x);
	      HELCoords[HELNumSommets+i].y=0.5*(HELCoords[oldnextid].y+HELCoords[currentsommet].y);
	      //on avance jusqu'au prochain point d'insertion
	      for (int j=0;j<insertevry+1;j++) currentsommet= ElSommets[currentsommet].NextId;
	      //il y a un "+1" car il faut tenir compte du sommet qu'on vient d'inserer
	    }
	  unsigned long lefttoadd=stoadd; //nbre de sommets restant a ajouter
	  //on regarde si on peut rajouter des threads
	  if (ElThreads<ElMaxThreads)
	    {
	      unsigned long ttoadd=ElMaxThreads-ElThreads;
	      if (ttoadd>stoadd) ttoadd=stoadd;
	      if (Output>=UltraDetailed) printf("HElastique: ajout de %d threads\n",ttoadd);
	      //ajout des threads,
	      //si on peut en rajouter c'est que ElSPerT=1 et ElFirstFavThread=0
	      //on rajoute les threads, en attribuant a chacun un sommet
	      for (int i=0;i<ttoadd;i++)
		{
		  ElThreadFirstSommet[ElThreads+i]=HELNumSommets+i;
		  ElSommets[HELNumSommets+i].NextSommet=NULL;
		  pthread_t inu;
		  pthread_create(&inu,NULL,HEL_WorkingThread,&ElSommets[ElThreadFirstSommet[ElThreads+i]]);
		}
	      //il reste stoadd-ttoadd sommets a ajouter
	      lefttoadd=stoadd-ttoadd;
	      ElThreads+=ttoadd;
	    }
	  //on repartis  les lefttoadd sommets , on en donne 1 par thread a partir de ElFirstFavThread
	  unsigned long firsttoadd=HELNumSommets+stoadd-lefttoadd;
	  for (int i=0;i<lefttoadd;i++)
	    {
	      //on s'insere en deuxieme position
	      unsigned long fs=ElThreadFirstSommet[ElFirstFavThread];
	      SommetInfo *ss=ElSommets[fs].NextSommet;
	      ElSommets[fs].NextSommet=&ElSommets[firsttoadd+i];
	      ElSommets[firsttoadd+i].NextSommet=ss;
	      ElFirstFavThread++;
	      if (ElFirstFavThread==ElThreads)
		{
		  ElFirstFavThread=0;
		  ElSPerT++;
		}
	    }
	  //on rerepartis les HandleSum entre les sommets
	  unsigned long lastshandlecount=0;//nbre de handlesum du dernier sommet
	  unsigned long nextsinc=HELNumSommets; //le prochain auquel rajouter un sumhandle
	  while (lastshandlecount<ElHPerS)
	    {
	      //on deplace le dernier handlesum du dernier defavorise
	      int firstdefav=ElFirstFavSommet-1;
	      if (firstdefav==-1)
		{
		  firstdefav=HELNumSommets-1;
		  ElHPerS--;
		}
	      ElSommets[nextsinc].HandleSum[lastshandlecount]=ElSommets[firstdefav].HandleSum[ElHPerS];
	      //ligne prec ElHPerS+1-1: +1 parce qu'il est defavorise -1 parce que 0-based
	      ElSommets[firstdefav].HandleSum[ElHPerS]=-1;
	      ElSommets[nextsinc].HandleSum[lastshandlecount+1]=-1;
	      ElFirstFavSommet=ElFirstFavSommet?ElFirstFavSommet-1:HELNumSommets-1;
	      nextsinc++;
	      if (nextsinc==HELNumSommets+stoadd)
		{
		  nextsinc=HELNumSommets;
		  lastshandlecount++;
		}
	    }
	  //fin de repartition des handlesums
	  HELNumSommets+=stoadd;
	}
    }//fin boucle sur HELK
  //on tue tous les threads
  HELState=Die;
  synchronizer_signal(HELSem);
  //il reste plus qu'a deduire la solution de l'elastique
  //on attribue a chaque ville le sommet de l'elastique dont il est le plus proche
  //on stoque cette info dans les handlesum
  for (int i=0;i<HELNumSommets;i++) ElSommets[i].HandleSum[0]=-1;
  for (int i=0;i<e->GetSommets();i++)
    {
      double mindist=e->GetMaxCoord()*e->GetMaxCoord()*763; //plus grand que toute distance entre deux points
      int bests;
      Point2D* c=e->GetCoords(i);
      for (int j=0;j<HELNumSommets;j++)
	{
	  double dist= ( ((double)c->x)- HELCoords[j].x) * ( ((double)c->x)- HELCoords[j].x)
	    + ( ((double)c->y)- HELCoords[j].y) * ( ((double)c->y)- HELCoords[j].y);
	  if (dist<mindist)
	    {
	      mindist=dist;
	      bests=j;
	    }
	}
      while (1)
	{
	  int pos=0;
	  for (;(pos>5)||(ElSommets[bests].HandleSum[pos]!=-1);pos++);
	  if (pos>5)
	    {
	      //oula
	      if (HELOutput>=Important) printf("HElastique: trop de villes proche du sommet %d\n",bests);
	      bests=ElSommets[bests].NextId;
	      continue;
	    }
	  ElSommets[bests].HandleSum[pos]=i;
	  if (pos<5)  ElSommets[bests].HandleSum[pos+1]=-1; 
	  break; 
	}
    }
  //on construit le tArretes fort de ces informations
  tArretes  * ta=new tArretes(e->GetSommets());
  int currentsommet=0;
  int lastville=-1;
  int firstville;
  for (;(lastville==-1)||(currentsommet!=0);currentsommet=ElSommets[currentsommet].NextId)
    {
      for (int i=0;(i<6)&&(ElSommets[currentsommet].HandleSum[i]!=-1);i++)
	{
	  if (lastville==-1)
	    {
	      lastville=ElSommets[currentsommet].HandleSum[i];
	      firstville=lastville;
	      continue;
	    }
	  ta->Add(lastville,ElSommets[currentsommet].HandleSum[i]);
	  lastville=ElSommets[currentsommet].HandleSum[i];
	  continue;
	}
    }
  ta->Add(lastville,firstville);
  //on libere toute la memoire allouee
  free(ElSommets);
  free(HELWeightSum);
  for( int i=0;i<e->GetSommets();i++) free(HELWeights[i]);
  free(HELWeights);
  free(HELCoords);
  free(ElThreadFirstSommet);
  synchronizer_destroy(HELSem);
  if (Output>=Important) printf("HElastique: chemin de longueur %d trouv\n",e->Eval(ta));
  return ta;
}




/**************************************************************************************/
PVC_Algo  Algorithmes[]=
  {
    {"HPlusProcheVoisin","Insertion a chaque etape du sommet le plus proche du dernier insere",PVC_HPlusProcheVoisin},
    {"HMoindreCout","Insertion a chaque etape du sommet produisant le circuit de moindre cout",PVC_HMoindreCout},
    {"H2Opt","Heuristique de voisinage 2-opt",PVC_H2Opt},
    {"HRecuitSimule","Heuristique du recuit simul, transitions 2-opt",PVC_HRecuitSimule},
    {"HARPM","Parcours prefixe d'un arbre recouvrant de poid minimal",PVC_HARPM},
    {"ASepEval","Algorithme donnant la solution optimale, methode de sparation/valuation",PVC_ASeparationEvaluation},
    {"HElastique","Heuristique de l'elastique",PVC_HElastique},
    {NULL,NULL,NULL}
  };
