//graphe.cpp
#include "graphe.h"
#include <memory.h>
#include <unistd.h>
#include <alloc.h>
#include <stdlib.h>
#include <stdio.h>
//#include <GL/gl.h>
//#include <GL/glut.h>
#include "lc.hpp"
Graphe::Graphe(unsigned int s)
{
  Sommets=s;
  //allocation de la memoire, le minimum necessaire:la partie strictement au dessus de la diagonale de la matrice s par s
  Poids=(unsigned long **)malloc(sizeof(unsigned long *)*(Sommets-1));
  for (int i=0;i<Sommets-1;i++)
    Poids[i]=(unsigned long *)malloc(sizeof(unsigned long)*(Sommets-1-i));
}

unsigned long  Graphe::Get(int s1,int s2)
{
  //conversion sommets=>indices dans le tableau
  if (s1>s2)
    {
      s1=s1^s2;
      s2=s1^s2;
      s1=s1^s2;
    }
  if ((s2>=Sommets)||(s1==s2)) return -1;
  return Poids[s1][s2-s1-1];
}

void Graphe::Set(int s1,int s2,unsigned long  p)
{
  //conversion sommets=>indices dans le tableau
  if (s1>s2)
    {
      s1=s1^s2;
      s2=s1^s2;
      s1=s1^s2;
    }
  if ((s2>=Sommets)||(s1==s2)) return ;
  Poids[s1][s2-s1-1]=p;
}

unsigned int  Graphe::Eval(tArretes *a)
{
  tArrete * t=NULL;
  unsigned int Length=0;
  while(a->Get(&t))
	Length+=Get(t->s1,t->s2);
  return Length;
}

tArretes::tArretes(unsigned int s)
{
  Sommets=s;
  Pointers=(tArrete**)malloc(sizeof(tArrete *)*(s-1));//inutile pour le dernier sommmet
  Pointers2=(tArrete**)malloc(sizeof(tArrete *)*(s-1)); //inutile pour le premier sommet
  memset(Pointers,0,sizeof(tArrete *)*(s-1));
  memset(Pointers2,0,sizeof(tArrete *)*(s-1));
}

tArretes::tArretes(tArretes *src)
{
  Sommets=src->Sommets;
  Pointers=(tArrete**)malloc(sizeof(tArrete *)*(Sommets-1));//inutile pour le dernier sommmet
  Pointers2=(tArrete**)malloc(sizeof(tArrete *)*(Sommets-1)); //inutile pour le premier sommet
  memset(Pointers,0,sizeof(tArrete *)*(Sommets-1));
  memset(Pointers2,0,sizeof(tArrete *)*(Sommets-1));
  tArrete * ar=NULL;
  while (src->Get(&ar)) Add(ar->s1,ar->s2);
}

bool tArretes::Present(int s1,int s2)
{
  if (s1>s2)
    {
      s1=s1^s2;
      s2=s1^s2;
      s1=s1^s2;
    }
  if (s1==s2) return true;
  if (s2>=Sommets) return false;
  for (tArrete * a=Pointers[s1];a&&(a->s2<=s2);a=(tArrete *)a->GetNext())
    if (a->s2==s2) return true;
  return false;
}

bool tArretes::Add(int s1,int s2)
{
  if (s1>s2)
    {
      s1=s1^s2;
      s2=s1^s2;
      s1=s1^s2;
    }
  if (s1==s2) return false;  
  if (s2>=Sommets) return false;
  tArrete * a; //l'arrete qui sera creee
  //maj liste sur s1
  if (!Pointers[s1])
    {
      a=new tArrete();
      a->s1=s1;
      a->s2=s2;
      Pointers[s1]=a;
    }
  else
    {
      //on parcours jusqu'a trouver le point d'insertion
      tArrete * t,*prev;
      for (t=Pointers[s1],prev=t;t;prev=t,t=(tArrete *)t->GetNext())
	{
	  if (t->s2>=s2) break;
	}
      if ((t)&&(t->s2==s2)) return false; //deja present
      //on insere juste apres prev sauf si t==prev
      a=new tArrete();
      a->s1=s1;
      a->s2=s2;
      if (t==prev)
	{
	  prev->InsertElementBefore(a);
	  Pointers[s1]=a;	
	}
      else
	prev->InsertElementAfter(a);
    }
  //maj liste sur s2
  if (!Pointers2[s2-1]) Pointers2[s2-1]=a;
  else
    {//recherche du point d'insertion
      for (tArrete *t=Pointers2[s2-1];;t=t->GetNext2())
	{
	  //un des deux cas suivant finira necessairement par etre vrai
	  if (t->s1>a->s1)
	    {//insertion avant
	      t->InsertBefore2(a);
	      if (Pointers2[s2-1]==t) Pointers2[s2-1]=a;
	      return true;
	    }
	  if (!t->GetNext2())
	    {
	      t->InsertAtEnd2(a);
	      return true;
	    }
	}
    }
  return true;
}

void tArretes::Clear()
{
  for (int i=0;i<Sommets-1;i++)
    {
      EradicateLC<tArrete>(Pointers[i]);
      Pointers[i]=0;
      Pointers2[i]=0;
    }
}

bool tArretes::Remove(int s1,int s2)
{ 
  if (s1>s2)
    {
      s1=s1^s2;
      s2=s1^s2;
      s1=s1^s2;
    }
  if (s1==s2) return false;  
  if (s2>=Sommets) return false;
  if (!Pointers[s1]) return false;
  for (tArrete *t=Pointers[s1];(t)&&(t->s2<=s2);t=(tArrete *)t->GetNext())
    {
      if (t->s2==s2)
	{
	  if (t==Pointers[s1]) Pointers[s1]=(tArrete *)t->GetNext();
	  if (t==Pointers2[s2-1]) Pointers2[s2-1]=t->GetNext2();
	  t->EradicateFromLC();
	  t->EradicateFromLC2();
	  delete t;
	  return true;
	}
    }
  return false;
}

bool tArretes::Boucle(int s1,int s2)
{//l'algorithme de detection de boucle incluant une arrete donnBBie est assez simple
  //on part de s1, on effectue un parcours du graphe par niveau jusqu'a atteindre s2 ou avoir essaye toutes les arretes
  char *hit=new char[Sommets]; //liste des sommets deja traites
  //on utilise une pile pour avoir la liste des sommets nouvellement atteints
  char *stack[2];
  stack[0]=new char[Sommets];  
  stack[1]=new char[Sommets];
  unsigned int stackpos[2]; //premie element libre de la stack
  unsigned int CurrentStack=0;
  memset(hit,0,Sommets);
  stackpos[0]=1;
  stack[0][0]=s1;
  while (stackpos[CurrentStack])
    {
      //on parcours les arretes accessibles de stackpos[currentstack]
      //on remplit stackpos[1-CurrentStack] avec les nouveaux sommets atteints;
      stackpos[1-CurrentStack]=0;
      for (int i=0;i<stackpos[CurrentStack];i++)
	{
	  //on enumere les arretes passant par ce sommet
	  tArrete * ar=NULL;
	  int s;
	  while((s=Get(&ar,stack[CurrentStack][i]))!=-1)
	    {
	      if (s==s2)
		{
		  delete stack[0];
		  delete stack[1];
		  delete hit;
		  return true;
		}
	      if (hit[s]) continue; //deja rajoute
	      hit[s]=1;
	      stack[1-CurrentStack][stackpos[1-CurrentStack]++]=s;
	    }
	}
      CurrentStack=1-CurrentStack; //on inverse
    }
  delete hit;
  delete stack[0];
  delete stack[1];
  return false;
}

bool tArretes::Get(tArrete **a)
{
  if (!*a)
    {//premiere arrete de l'enumeration
      int i;
      for (i=0;(i<Sommets-1)&&(!Pointers[i]);i++) ;
      if (i>=Sommets-1) return false; //aucune arrete
      *a=Pointers[i];
      return true;
    }
  if ((*a)->GetNext())
    {
      *a=(tArrete *)(*a)->GetNext();
      return true;
    }
  //on passe au 's1' suivant
  int i;
  for (i=(*a)->s1+1;(i<Sommets-1)&&(!Pointers[i]);i++) ;
  if (i>=Sommets-1) return false; //aucune arrete
  *a=Pointers[i];
  return true;
}

int tArretes::Get(tArrete **a,int s1)
{
  if (!(*a))
    {
      //attention au cas particulier s1==Somets, car Pointers[Sommets] n'existe pas
      if (s1==Sommets-1)
	{
	  *a=Pointers2[Sommets-2];
	  return (*a)?(*a)->s1:-1;
	} 
      *a=Pointers[s1];
      if (*a) return (*a)->s2;
      if (s1==0) return -1;
      *a=Pointers2[s1-1];
      if (!(*a)) return -1;
      return (*a)->s1;
    }
  if ((*a)->s1==s1)
    {//on est en phase un
      if ((*a)->GetNext())
	{
	  *a=(tArrete *)(*a)->GetNext();
	  return (*a)->s2;
	}
      else
	{//on passe en phase deux
	  //la encore attention, Pointers2 n'a pas d'entree pour le sommet 0
	  if (s1==0) return -1;	  
	  *a=Pointers2[s1-1];
	  if (*a) return (*a)->s1;
	  else return -1;
	}
    }
  //on est en phase deux (else implicite)
  *a=(*a)->GetNext2();
  if (*a) return (*a)->s1;
  return -1;
}

tArrete * tArretes::CircuitAsLinkedList()
{
  tArrete *first=new tArrete();
  first->s1=0;
  /*
    printf("tArretes dumping\n");
    for (int i=0;i<Sommets;i++)
    {
    tArrete * dbg=NULL;
    int ds2;
    while (1) 
    {
    ds2=Get(&dbg,i);
    if (ds2==-1) break;
    printf("- %d-%d \n",i,ds2);
    }
    if (i!=0)
    {
    for (dbg=Pointers2[i-1];dbg;dbg=dbg->GetNext2())
    printf("+ %d-%d \n",dbg->s2,dbg->s1);
    }
    }*/
  tArrete *a=Pointers[0];
  if (!a) return NULL;
  first->s2=a->s2;
  tArrete *last=first;
  for (int i=2;i<Sommets;i++)
    {
      a=NULL;
      int nexts;
      do {nexts=Get(&a,last->s2);} while ((nexts!=-1)&&(nexts==last->s1));
      if (nexts==-1) return NULL;
      tArrete * nexta=new tArrete();
      nexta->s1=last->s2;
      nexta->s2=nexts;
      last->InsertAtEnd2(nexta);
      last=nexta;
    }
  //reste a boucler
  tArrete *boucle=new tArrete();
  boucle->s1=last->s2;
  boucle->s2=first->s1;
  last->SetNext2(boucle);
  boucle->SetPrev2(last);
  boucle->SetNext2(first);
  first->SetPrev2(boucle);
  return first;
}

int * tArretes::CircuitAsArray()
{
  int * c=(int *)malloc(sizeof(int)*(Sommets+1));
  c[0]=0;
  c[Sommets]=0;// perdre quatre octes est negligeable devant la simplification que cela apporte au niveau du code: un cas particulier en moins a traiter
  tArrete *a=Pointers[0];
  if (!a) return NULL;
  c[1]=a->s2;
  for (int i=2;i<Sommets;i++)
    {
      a=NULL;
      int nexts;
      do {nexts=Get(&a,c[i-1]);} while ((nexts!=-1)&&(nexts==c[i-2]));
      if (nexts==-1) return NULL;
      c[i]=nexts;
    }
  return c;
}

EuclidianGraphe::EuclidianGraphe(int Sommets)
  :Graphe(Sommets)
{
  printf("%d\n",Sommets);
  Coordinates=(Point2D *)malloc(sizeof(Point2D)*Sommets);
  DrawingReady=false;
  DeltaCoordsAvailable=false;
}

void EuclidianGraphe::Randomise(unsigned long Resolution,unsigned long ScaleFactor)
{
  MaxCoord=Resolution*ScaleFactor;
  DeltaCoordsAvailable=false;
  WorldSize=Resolution*ScaleFactor;
  srand(time(NULL));
  //generation des coordonnees
  printf("Coordonnees des sommets:\n");
  for (int i=0;i<Sommets;i++)
    {
      float t=((float)rand())/((float)RAND_MAX)*((float)Resolution);
      Coordinates[i].x=((int)t)*ScaleFactor;
      t=((float)rand())/((float)RAND_MAX)*((float)Resolution);
      Coordinates[i].y=((int)t)*ScaleFactor;
      printf("%d: %u,%u\n",i,Coordinates[i].x,Coordinates[i].y);
    }
  //generation des poids
  for (int i=0;i<Sommets;i++)
    {
      for (int j=i+1;j<Sommets;j++)
	{
	  float dx=(long)(Coordinates[i].x-Coordinates[j].x);
	  float dy=(long)(Coordinates[i].y-Coordinates[j].y);
	  float d=sqrt(dx*dx+dy*dy);
	  Set(i,j,(unsigned long)d);
	}
    }
#ifdef DEBUG //c'est un peu long a afficher pour 500 villes...
  printf("\nMatrice des poids:\n");
  for (int i=0;i<Sommets;i++)
    {
      for (int j=0;j<Sommets;j++)
	{
	  printf("%d ",Get(i,j));
	}
      printf("\n");
    }
#endif
}

unsigned long EuclidianGraphe::WindowX=0;
unsigned long EuclidianGraphe::WindowY=0;
void DrawChangeSize(int x,int y)
{
  EuclidianGraphe::WindowX=x;
  EuclidianGraphe::WindowY=y;
}
bool EuclidianGraphe::DrawingReady=false;
int EuclidianGraphe::WindowId=0;
EuclidianGraphe * DrawMe=NULL;
tArretes * DrawMeToo=NULL;
fDraw DrawMeAlso=NULL;
void DrawDraw()
{
  if (DrawMe)
    {
      //printf("COOLdrawdraw\n");
      DrawMe->DoDraw(DrawMeToo);
      DrawMe=NULL;
      return;
    }  
  //printf("drawdraw\n");
  //  glClear(GL_COLOR_BUFFER_BIT);
  //glutSwapBuffers();
}
void DrawTestDraw()
{
  if (DrawMe) DrawDraw();
}
void * glutstart(void *p)
{
  /*
  void **q=(void **)p;

 glutInit((int *)q[0],* ((char ***)q[1]));
  glutInitDisplayMode (GLUT_DOUBLE|GLUT_RGB);
  glutInitWindowSize (400, 400);
  EuclidianGraphe::WindowX=400;
  EuclidianGraphe::WindowY=400;
  glutInitWindowPosition (50, 50);
  EuclidianGraphe::WindowId=glutCreateWindow ("PVC");
  glutDisplayFunc(DrawDraw);
  glutReshapeFunc(DrawChangeSize);
  glutIdleFunc(DrawTestDraw);
  GLfloat specular [] = { .5, .5, .5, 1.0 };
  GLfloat ambient [] = { .2, .2, .2, 1.0 };
  GLfloat diffuse [] = { .8, .8, .8, 1.0 };
  GLfloat shininess [] = { 127.0 };
  GLfloat position [] = { 80.0, 80.0, 25.0, 1.0 };
 // Set the shading model
  glShadeModel(GL_SMOOTH);
 // Set the polygon mode to fill
 glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
 glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
 glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
 glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
 glEnable(GL_COLOR_MATERIAL);
 glLightfv(GL_LIGHT0,GL_SPECULAR,specular);
 glLightfv(GL_LIGHT0,GL_AMBIENT,ambient);
 glLightfv(GL_LIGHT0,GL_DIFFUSE,diffuse);
 glLightfv(GL_LIGHT0, GL_POSITION, position);
 //glEnable(GL_LIGHTING);
 //glEnable(GL_LIGHT0);
glutMainLoop();
  */
}
void EuclidianGraphe::InitDrawing(int pargc,char * argv[])
{
  /*
  if (DrawingReady) return;
  //initialisation OpenGL/Glut
  void * p[2]={&pargc,&argv};
pthread_t *pt=new pthread_t();
 pthread_create(pt,NULL,(void *(*)(void *))glutstart,p);
 sleep(2);
  */
}

void EuclidianGraphe::Draw(tArretes *a,fDraw  func)
{
  printf("Drawing disabled in this special edition\n");
  /*
  DrawMe=this;
  DrawMeToo=a;
  DrawMeAlso=func;
  glutSetWindow(WindowId);
  for (int i=0;(i<50)&&DrawMe;i++) usleep(4000);
  if (DrawMe) printf("failed draw\n");
  */
}
void EuclidianGraphe::DoDraw(tArretes *a)
{
  /*
  if (!DeltaCoordsAvailable)
    {
      DeltaCoordsAvailable=true;
      int a=WorldSize/100; //taille des rectangles
      DeltaCoords[1][1]=DeltaCoords[2][1]=a/2; //y
      DeltaCoords[0][1]=a*(-1);
      DeltaCoords[0][0]=0; //deltax du sommet
      DeltaCoords[2][0]=((float)a*sqrt(3.0f)/2.0f);
      DeltaCoords[1][0]=(-1)*DeltaCoords[2][0];
    }
  glViewport(0,0,WindowX,WindowY);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0,WorldSize,0,WorldSize);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glClear(GL_COLOR_BUFFER_BIT);
  //on dessine les villes, sous la forme de petits triangles equilateraux
  glBegin(GL_TRIANGLES);
  glColor3f(1,1,1);
  for (int i=0;i<Sommets;i++)
    {
      glColor3f(1,(float)i/(float)Sommets,1-(float)i/(float)Sommets);
      glVertex2i(Coordinates[i].x+DeltaCoords[0][0],Coordinates[i].y+DeltaCoords[0][1]);
      glVertex2i(Coordinates[i].x+DeltaCoords[1][0],Coordinates[i].y+DeltaCoords[1][1]);
      glVertex2i(Coordinates[i].x+DeltaCoords[2][0],Coordinates[i].y+DeltaCoords[2][1]);
      //printf("%d %d %d\n",i,Coordinates[i].x,Coordinates[i].y);
    }
  glEnd();
  //on dessine les arretes
  if (!a) 
    {
      if (DrawMeAlso) DrawMeAlso();
      glutSwapBuffers();
      return;
    }
  tArrete *ar=NULL;
  glBegin(GL_LINES);
  glColor3f(1,1,1);
  while (a->Get(&ar))
    {
      //printf ("%d <-> %d\n",ar->s1,ar->s2);
      glVertex2i(Coordinates[ar->s1].x,Coordinates[ar->s1].y);
      glVertex2i(Coordinates[ar->s2].x,Coordinates[ar->s2].y);
    }
  glEnd();
  if (DrawMeAlso) DrawMeAlso();
  glutSwapBuffers();
  */
}
