//graphe.h
#ifndef GRAPHE_H
#define GRAPHE_H
#include "lc.h"
class tArretes;
class Graphe
{
 protected:
  unsigned long ** Poids;
  unsigned int Sommets;
 public:
  unsigned long  Get(int,int); //renvoit un poid
  void Set(int,int,unsigned long); //assigne un poid
  Graphe(unsigned int Sommets);
  unsigned int GetSommets() {return Sommets;}
  unsigned int Eval(tArretes * a);
};

struct Point2D
{
  unsigned long x;
  unsigned long y;
};

class tArretes;
typedef void(*fDraw)(void); //draw callback
class EuclidianGraphe: public Graphe
{
  Point2D  * Coordinates;
  static bool DrawingReady;
  unsigned long  WorldSize; //tous les points sont entre (0,0) et (WorldSize,WorldSize)
  unsigned long DeltaCoords[3][2];//cooronnees precalcules des sommets des triangles
  bool DeltaCoordsAvailable; //calcul effectue?
  unsigned long MaxCoord; //Resolution*Scalefactor si randomise
 public:
  unsigned long GetMaxCoord() {return MaxCoord;}
  static int WindowId;
  Point2D * GetCoords(int p) {return &Coordinates[p];}
  static unsigned long WindowX,WindowY; //mis a jour dans une fonction globale
  EuclidianGraphe(int Sommets);
  void Randomise(unsigned long Resolution,unsigned long ScaleFactor); //initialisation des coordonnees des points aleatoires, cf doc pour parametres
  //void InitPoids(char * buffer, eFormat Format); //initialise les poids depuis des donnees TSP
  //void InitDists(char * buffer); //initialise les distances depuis des donnees TSP 
  //static EuclidianGraphe LoadTSP(char * FileName);
  void Draw(tArretes * a,fDraw func=NULL); //affiche les sommets et le sousensemble d'arretes a
  void DoDraw(tArretes *a); //probleme de syncronisation
  static void InitDrawing(int pargc,char * argv[]); //initialisation opengl
};

class tArrete:public ListeChainee
{
  //Next et Prev de ListeChainee font le chainage selon s1
  tArrete * Prev2;
  tArrete * Next2; //chainage selon s2;
 public:
  //fonctions de la deuxieme liste chainee
  void EradicateFromLC2() {if (Prev2) Prev2->Next2=Next2;if (Next2) Next2->Prev2=Prev2; Prev2=NULL;Next2=NULL;}
  tArrete * GetNext2() {return Next2;}
  void SetNext2(tArrete *n) {Next2=n;}
void SetPrev2(tArrete *n) {Prev2=n;}
  void InsertBefore2(tArrete *b) {b->Next2=this;b->Prev2=Prev2;if (Prev2) Prev2->Next2=b;Prev2=b;}
  void InsertAtEnd2(tArrete *b) {Next2=b;b->Prev2=this;}
  int s1,s2;
  int Tag; //utilisable par les  algo
  tArrete() {InitLC();Next2=NULL;Prev2=NULL;Tag=0;}
  //support inversion, cf doc H2Opt pour explication
  int GetS1(bool reversed) {return reversed?s2:s1;}
  int GetS2(bool reversed) {return reversed?s1:s2;} 
  void SetS1(int s,bool reversed) {if (reversed) s2=s; else s1=s;}
  void SetS2(int s,bool reversed) {if (reversed) s1=s; else s2=s;}
  tArrete * GetNext2(bool reversed) {return reversed?Prev2:Next2;}
  tArrete * GetPrev2(bool reversed) {return reversed?Next2:Prev2;}
  //versions avec maj du bool reversed, attention il faut partir ensuite de l'arrete renvoyee, ces fonctions ne sont pas idenpotentes
  tArrete * GetNext2(bool &reversed,char * revs) {tArrete * result=reversed?Prev2:Next2;if (revs[GetS2(reversed)]) reversed=!reversed;  return result;}
  tArrete * GetPrev2(bool &reversed,char * revs) {tArrete * result=reversed?Next2:Prev2;if (revs[GetS1(reversed)]) reversed=!reversed;  return result;}
  void SetNext2(tArrete *t,bool reversed) {if (reversed) Prev2=t; else Next2=t;}
  void SetPrev2(tArrete *t,bool reversed) {if (reversed) Next2=t; else Prev2=t;}
};

class tArretes //utilisee pour manipuler un ensemble d'arretes
{
  unsigned int Sommets; //nombre de sommets
  tArrete ** Pointers; //pour acces rapide, pointer vers le premier element avec s1==indice
  tArrete ** Pointers2; //chainage selon s2:premier element verifiant s2=index 
public:
  tArretes(unsigned int Sommets);
  tArretes(tArretes * src); //constructeur copie
  void Clear();
  bool Present(int s1,int s2); //renvoit true si le sommet est dans la liste
  bool Add(int s1,int s2); //renvoit false si deja present
  bool Remove(int s1,int s2); //renvoit false si non present
  bool Boucle(int s1,int s2); //ajouter {s1;s2} provoque une boucle?
  int Get(tArrete **a,int s1); //enumere les arretes contenant s1, renvoit s2 ou -1 si fin de l'enumeration, *a doit initialement etre NULL, puis etre passe a chaque appel
  bool Get(tArrete **a); //enumere toutes les adresses, mettre *a a NULL initialement renvoit false si plus d'arretes
  int * CircuitAsArray(); //si this definit un circuit renvoit les sommets dans l'ordre du parcours
  tArrete * CircuitAsLinkedList(); //circuit sous forme d'une l.c. circulaire (pointeurs Next2 et Prev2)
};

#endif



